/*
 * userspace pagecache management using LD_PRELOAD, fadvise and
 * sync_file_range().
 *
 * Andrew Morton <akpm@linux-foundation.org>
 * March, 2007
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <dlfcn.h>
#include <limits.h>

#include "sync_file_range.h"

enum fd_state {
	FDS_UNKNOWN = 0,	/* We know nothing about this fd */
	FDS_IGNORE,		/* Ignore this file (!S_ISREG?) */
	FDS_ACTIVE,		/* We're managing this file's pagecache */
};

struct fd_status {
	enum fd_state state;
	int was_written;
};

/*
 * These are set from the environment
 */
static unsigned long pagecache_max_bytes;
static unsigned long pagecache_chunk_size;
static int pagesize;

/*
 * Global state
 */
static unsigned long pagecache_size;

static struct fd_status *get_fd_status(int fd)
{
	static struct fd_status *fd_status;
	static int nr_fd_status;	/* Number at *fd_status */

	if (fd < 0)
		abort();
	if (fd + 1 > nr_fd_status) {
		fd_status = realloc(fd_status, sizeof(*fd_status) * (fd + 1));
		memset(fd_status + nr_fd_status, 0,
			sizeof(*fd_status) * (fd + 1 - nr_fd_status));
		nr_fd_status = fd + 1;
	}
	return fd_status;
}

static void grow_pagecache(int fd, size_t count)
{
	pagecache_size += count;
	if (pagecache_size > pagecache_max_bytes) {
		off_t off = lseek(fd, 0, SEEK_CUR);
		if (off > pagesize)
			posix_fadvise(fd, 0, off - pagesize,
					POSIX_FADV_DONTNEED);
		pagecache_size = 0;
	}
}

/*
 * Work out if we're interested in this fd
 */
static void inspect_fd(int fd, struct fd_status *fds)
{
	struct stat stat_buf;

	if (fstat(fd, &stat_buf))
		abort();
	fds->was_written = 0;
	if (S_ISREG(stat_buf.st_mode) || S_ISBLK(stat_buf.st_mode))
		fds->state = FDS_ACTIVE;
	else
		fds->state = FDS_IGNORE;
}

static void write_was_called(int fd, size_t count)
{
	struct fd_status *fds;

	if (pagecache_max_bytes == 0)
		return;

	fds = get_fd_status(fd);
	if (fds->state == FDS_UNKNOWN)
		inspect_fd(fd, fds);
	fds->was_written = 1;
	if (fds->state == FDS_IGNORE)
		return;
	grow_pagecache(fd, count);
}

static void read_was_called(int fd, size_t count)
{
	struct fd_status *fds;

	if (pagecache_max_bytes == 0)
		return;

	fds = get_fd_status(fd);
	if (fds->state == FDS_UNKNOWN)
		inspect_fd(fd, fds);
	if (fds->state == FDS_IGNORE)
		return;
	grow_pagecache(fd, count);
}

static void close_was_called(int fd)
{
	struct fd_status *fds;
	int err = 0;

	if (pagecache_max_bytes == 0)
		return;

	fds = get_fd_status(fd);
	if (fds->was_written)
		sync_file_range(fd, 0, LONG_MAX,
			SYNC_FILE_RANGE_WRITE|SYNC_FILE_RANGE_WAIT_AFTER);
	posix_fadvise(fd, 0, 0, POSIX_FADV_DONTNEED);
	fds->state = FDS_UNKNOWN;
}

static void parse_env(void)
{
	char *e;

	e = getenv("PAGECACHE_MAX_BYTES");
	if (e) {
		pagecache_max_bytes = strtoul(e, NULL, 10);
		pagecache_chunk_size = pagecache_max_bytes / 2;
		e = getenv("PAGECACHE_CHUNK_SIZE");
		if (e)
			pagecache_chunk_size = strtoul(e, NULL, 10);
	}
}

/*
 * syscall interface
 */

static ssize_t (*_write)(int fd, const void *buf, size_t count);
static ssize_t (*_pwrite)(int fd, const void *buf, size_t count, off_t offset);
static ssize_t (*_read)(int fd, void *buf, size_t count);
static ssize_t (*_pread)(int fd, void *buf, size_t count, off_t offset);
static int (*_close)(int fd);
static int (*_dup2)(int oldfd, int newfd);

static int symbols_loaded;

static void __load_symbols(void)
{
	void *handle;

	handle = dlopen("libc.so.6", RTLD_LAZY);
	if (!handle)
		abort();

	dlerror();
	_write = dlsym(handle, "write");
	if (dlerror())
		abort();

	_pwrite = dlsym(handle, "pwrite");
	if (dlerror())
		abort();

	dlerror();
	_read = dlsym(handle, "read");
	if (dlerror())
		abort();

	_pread = dlsym(handle, "pread");
	if (dlerror())
		abort();

	_close = dlsym(handle, "close");
	if (dlerror())
		abort();

	_dup2 = dlsym(handle, "dup2");
	if (dlerror())
		abort();

	dlclose(handle);

	parse_env();

	pagesize = getpagesize();

	symbols_loaded = 1;
}

static inline void load_symbols(void)
{
	if (!symbols_loaded)
		__load_symbols();
}

ssize_t write(int fd, const void *buf, size_t count)
{
	load_symbols();
	write_was_called(fd, count);
	return (*_write)(fd, buf, count);
}

ssize_t pwrite(int fd, const void *buf, size_t count, off_t offset)
{
	load_symbols();
	write_was_called(fd, count);
	return (*_pwrite)(fd, buf, count, offset);
}

ssize_t read(int fd, void *buf, size_t count)
{
	load_symbols();
	read_was_called(fd, count);
	return (*_read)(fd, buf, count);
}

ssize_t pread(int fd, void *buf, size_t count, off_t offset)
{
	load_symbols();
	read_was_called(fd, count);
	return (*_pread)(fd, buf, count, offset);
}

int close(int fd)
{
	load_symbols();
	close_was_called(fd);
	return (*_close)(fd);
}

int dup2(int oldfd, int newfd)
{
	load_symbols();
	close_was_called(oldfd);
	return (*_dup2)(oldfd, newfd);
}
